package sun.nio.ch;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.io.FileDescriptor;

public class IOUtil
{
    static final int IOV_MAX;
    
    private IOUtil() {
    }
    
    static int write(FileDescriptor fileDescriptor, ByteBuffer src, long n, NativeDispatcher nativeDispatcher) throws IOException {
        if ((src instanceof DirectBuffer)) {
            return writeFromNativeBuffer(fileDescriptor, src, n, nativeDispatcher);
        }
        int position = src.position();
        int limit = src.limit();
        assert position <= limit;
        ByteBuffer temporaryDirectBuffer = Util.getTemporaryDirectBuffer((position <= limit) ? (limit - position) : 0);
        try {
            temporaryDirectBuffer.put(src);
            temporaryDirectBuffer.flip();
            src.position(position);
            int writeFromNativeBuffer = writeFromNativeBuffer(fileDescriptor, temporaryDirectBuffer, n, nativeDispatcher);
            if (writeFromNativeBuffer > 0) {
                src.position(position + writeFromNativeBuffer);
            }
            return writeFromNativeBuffer;
        }
        finally {
            Util.offerFirstTemporaryDirectBuffer(temporaryDirectBuffer);
        }
    }
    
    private static int writeFromNativeBuffer(FileDescriptor fileDescriptor, ByteBuffer byteBuffer, long n, NativeDispatcher nativeDispatcher) throws IOException {
        int position = byteBuffer.position();
        int limit = byteBuffer.limit();
        assert position <= limit;
        int n2 = (position <= limit) ? (limit - position) : 0;
        if (n2 == 0) {
            return 0;
        }
        int n3;
        if (n != -1L) {
            n3 = nativeDispatcher.pwrite(fileDescriptor, ((DirectBuffer)byteBuffer).address() + (long)position, n2, n);
        }
        else {
            n3 = nativeDispatcher.write(fileDescriptor, ((DirectBuffer)byteBuffer).address() + (long)position, n2);
        }
        if (n3 > 0) {
            byteBuffer.position(position + n3);
        }
        return n3;
    }
    
    static long write(FileDescriptor fileDescriptor, ByteBuffer[] array, NativeDispatcher nativeDispatcher) throws IOException {
        return write(fileDescriptor, array, 0, array.length, nativeDispatcher);
    }
    
    static long write(FileDescriptor fileDescriptor, ByteBuffer[] array, int n, int n2, NativeDispatcher nativeDispatcher) throws IOException {
        IOVecWrapper value = IOVecWrapper.get(n2);
        boolean n3 = false;
        int n4 = 0;
        try {
            int n5 = n + n2;
            while (n < n5 && n4 < IOUtil.IOV_MAX) {
                ByteBuffer src = array[n];
                int n6 = src.position();
                int limit = src.limit();
                assert n6 <= limit;
                int n7 = (n6 <= limit) ? (limit - n6) : 0;
                if (n7 > 0) {
                    value.setBuffer(n4, src, n6, n7);
                    if (!(src instanceof DirectBuffer)) {
                        ByteBuffer temporaryDirectBuffer = Util.getTemporaryDirectBuffer(n7);
                        temporaryDirectBuffer.put(src);
                        temporaryDirectBuffer.flip();
                        value.setShadow(n4, temporaryDirectBuffer);
                        src.position(n6);
                        src = temporaryDirectBuffer;
                        n6 = temporaryDirectBuffer.position();
                    }
                    value.putBase(n4, ((DirectBuffer)src).address() + (long)n6);
                    value.putLen(n4, (long)n7);
                    ++n4;
                }
                ++n;
            }
            if (n4 == 0) {
                long n9 = 0L;
                if (n3 == 0) {
                    for (int k = 0; k < n4; ++k) {
                        ByteBuffer shadow3 = value.getShadow(k);
                        if (shadow3 != null) {
                            Util.offerLastTemporaryDirectBuffer(shadow3);
                        }
                        value.clearRefs(k);
                    }
                }
                return n9;
            }
            else {
                long writev = nativeDispatcher.writev(fileDescriptor, value.address, n4);
                for (int i = 0; i < n4; ++i) {
                    if (writev > 0L) {
                        ByteBuffer buffer = value.getBuffer(i);
                        int position = value.getPosition(i);
                        int remaining = value.getRemaining(i);
                        int n8 = (writev > (long)remaining) ? remaining : ((int)writev);
                        buffer.position(position + n8);
                        writev -= (long)n8;
                    }
                    ByteBuffer shadow = value.getShadow(i);
                    if (shadow != null) {
                        Util.offerLastTemporaryDirectBuffer(shadow);
                    }
                    value.clearRefs(i);
                }
                n3 = true;
                if (n3 == 0) {
                    for (int j = 0; j < n4; ++j) {
                        ByteBuffer shadow2 = value.getShadow(j);
                        if (shadow2 != null) {
                            Util.offerLastTemporaryDirectBuffer(shadow2);
                        }
                        value.clearRefs(j);
                    }
                }
                return writev;
            }
        }
        finally {
            if (n3 == 0) {
                for (int l = 0; l < n4; ++l) {
                    ByteBuffer shadow4 = value.getShadow(l);
                    if (shadow4 != null) {
                        Util.offerLastTemporaryDirectBuffer(shadow4);
                    }
                    value.clearRefs(l);
                }
            }
        }
    }
    
    static int read(FileDescriptor fileDescriptor, ByteBuffer byteBuffer, long n, NativeDispatcher nativeDispatcher) throws IOException {
        if (byteBuffer.isReadOnly()) {
            throw new IllegalArgumentException("Read-only buffer");
        }
        if ((byteBuffer instanceof DirectBuffer)) {
            return readIntoNativeBuffer(fileDescriptor, byteBuffer, n, nativeDispatcher);
        }
        ByteBuffer temporaryDirectBuffer = Util.getTemporaryDirectBuffer(byteBuffer.remaining());
        try {
            int intoNativeBuffer = readIntoNativeBuffer(fileDescriptor, temporaryDirectBuffer, n, nativeDispatcher);
            temporaryDirectBuffer.flip();
            if (intoNativeBuffer > 0) {
                byteBuffer.put(temporaryDirectBuffer);
            }
            return intoNativeBuffer;
        }
        finally {
            Util.offerFirstTemporaryDirectBuffer(temporaryDirectBuffer);
        }
    }
    
    private static int readIntoNativeBuffer(FileDescriptor fileDescriptor, ByteBuffer byteBuffer, long n, NativeDispatcher nativeDispatcher) throws IOException {
        int position = byteBuffer.position();
        int limit = byteBuffer.limit();
        assert position <= limit;
        int n2 = (position <= limit) ? (limit - position) : 0;
        if (n2 == 0) {
            return 0;
        }
        int n3;
        if (n != -1L) {
            n3 = nativeDispatcher.pread(fileDescriptor, ((DirectBuffer)byteBuffer).address() + (long)position, n2, n);
        }
        else {
            n3 = nativeDispatcher.read(fileDescriptor, ((DirectBuffer)byteBuffer).address() + (long)position, n2);
        }
        if (n3 > 0) {
            byteBuffer.position(position + n3);
        }
        return n3;
    }
    
    static long read(FileDescriptor fileDescriptor, ByteBuffer[] array, NativeDispatcher nativeDispatcher) throws IOException {
        return read(fileDescriptor, array, 0, array.length, nativeDispatcher);
    }
    
    static long read(FileDescriptor fileDescriptor, ByteBuffer[] array, int n, int n2, NativeDispatcher nativeDispatcher) throws IOException {
        IOVecWrapper value = IOVecWrapper.get(n2);
        boolean n3 = false;
        int n4 = 0;
        try {
            int n5 = n + n2;
            while (n < n5 && n4 < IOUtil.IOV_MAX) {
                ByteBuffer byteBuffer = array[n];
                if (byteBuffer.isReadOnly()) {
                    throw new IllegalArgumentException("Read-only buffer");
                }
                int n6 = byteBuffer.position();
                int limit = byteBuffer.limit();
                assert n6 <= limit;
                int n7 = (n6 <= limit) ? (limit - n6) : 0;
                if (n7 > 0) {
                    value.setBuffer(n4, byteBuffer, n6, n7);
                    if (!(byteBuffer instanceof DirectBuffer)) {
                        ByteBuffer temporaryDirectBuffer = Util.getTemporaryDirectBuffer(n7);
                        value.setShadow(n4, temporaryDirectBuffer);
                        byteBuffer = temporaryDirectBuffer;
                        n6 = temporaryDirectBuffer.position();
                    }
                    value.putBase(n4, ((DirectBuffer)byteBuffer).address() + (long)n6);
                    value.putLen(n4, (long)n7);
                    ++n4;
                }
                ++n;
            }
            if (n4 == 0) {
                long n9 = 0L;
                if (n3 == 0) {
                    for (int k = 0; k < n4; ++k) {
                        ByteBuffer shadow3 = value.getShadow(k);
                        if (shadow3 != null) {
                            Util.offerLastTemporaryDirectBuffer(shadow3);
                        }
                        value.clearRefs(k);
                    }
                }
                return n9;
            }
            else {
                long readv = nativeDispatcher.readv(fileDescriptor, value.address, n4);
                for (int i = 0; i < n4; ++i) {
                    ByteBuffer shadow = value.getShadow(i);
                    if (readv > 0L) {
                        ByteBuffer buffer = value.getBuffer(i);
                        int remaining = value.getRemaining(i);
                        int n8 = (readv > (long)remaining) ? remaining : ((int)readv);
                        if (shadow == null) {
                            buffer.position(value.getPosition(i) + n8);
                        }
                        else {
                            shadow.limit(shadow.position() + n8);
                            buffer.put(shadow);
                        }
                        readv -= (long)n8;
                    }
                    if (shadow != null) {
                        Util.offerLastTemporaryDirectBuffer(shadow);
                    }
                    value.clearRefs(i);
                }
                n3 = true;
                if (n3 == 0) {
                    for (int j = 0; j < n4; ++j) {
                        ByteBuffer shadow2 = value.getShadow(j);
                        if (shadow2 != null) {
                            Util.offerLastTemporaryDirectBuffer(shadow2);
                        }
                        value.clearRefs(j);
                    }
                }
                return readv;
            }
        }
        finally {
            if (n3 == 0) {
                for (int l = 0; l < n4; ++l) {
                    ByteBuffer shadow4 = value.getShadow(l);
                    if (shadow4 != null) {
                        Util.offerLastTemporaryDirectBuffer(shadow4);
                    }
                    value.clearRefs(l);
                }
            }
        }
    }
    
    public static FileDescriptor newFD(int n) {
        FileDescriptor fileDescriptor = new FileDescriptor();
        setfdVal(fileDescriptor, n);
        return fileDescriptor;
    }
    
    static native boolean randomBytes(byte[] p0);
    
    static native long makePipe(boolean p0);
    
    static native boolean drain(int p0) throws IOException;
    
    public static native void configureBlocking(FileDescriptor p0, boolean p1) throws IOException;
    
    public static native int fdVal(FileDescriptor p0);
    
    static native void setfdVal(FileDescriptor p0, int p1);
    
    static native int fdLimit();
    
    static native int iovMax();
    
    static native void initIDs();
    
    public static void load() {
    }
    
    static {
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                System.loadLibrary("net");
                System.loadLibrary("nio");
                return null;
            }
        });
        initIDs();
        IOV_MAX = iovMax();
    }
}
